/*
								+----------------------------------+
								|                                  |
								| ***  Special shader sources  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|          SpecialShader.h         |
								|                                  |
								+----------------------------------+
*/

#ifndef __SPECIALIZED_SHADERS_INCLUDED
#define __SPECIALIZED_SHADERS_INCLUDED

const TShaderInfo t_perpixel_equalize("per-pixel_equalize_fs", TShaderInfo::proc_Fragment,
	"uniform sampler2D n_image, n_minima, n_maxima;\n"
	"uniform float f_brightness, f_headroom, f_equalize, f_desaturate;\n"
	"\n"
	"void main()\n"
	"{\n"
	"    vec3 v_min = texture2D(n_minima, gl_TexCoord[0].xy).xyz;\n"
	"    vec3 v_max = texture2D(n_maxima, gl_TexCoord[0].xy).xyz;\n"
	"    v_max -= v_min;\n"
	"    v_max += vec3(f_headroom);\n"
	"    vec3 v_image = texture2D(n_image, gl_TexCoord[0].xy).xyz;\n"
	"    vec3 v_color = vec3(f_brightness) + (v_image - v_min * f_equalize) / v_max;\n"
	"    vec3 v_grey = vec3(dot(v_color, vec3(.299, .587, .114)));\n"
	"    gl_FragColor = vec4(v_color + f_desaturate * (v_grey - v_color), 1.0);\n"
	"}\n", "n_image|n_minima|n_maxima");

const TShaderInfo t_perpixel_equalize_mono("monochrome_per-pixel_equalize_fs", TShaderInfo::proc_Fragment,
	"uniform sampler2D n_image, n_minima, n_maxima;\n"
	"uniform float f_brightness, f_headroom, f_equalize, f_desaturate;\n"
	"\n"
	"void main()\n"
	"{\n"
	"    vec3 v_min = vec3(dot(texture2D(n_minima, gl_TexCoord[0].xy).xyz, vec3(.299, .587, .114)));\n" // .299, .587, .114
	"    vec3 v_max = vec3(dot(texture2D(n_maxima, gl_TexCoord[0].xy).xyz, vec3(.299, .587, .114)));\n" // enhance blaw contrast
	"    v_max -= v_min;\n"
	"    v_max += vec3(f_headroom);\n"
	"    vec3 v_image = texture2D(n_image, gl_TexCoord[0].xy).xyz;\n"
	"    vec3 v_color = vec3(f_brightness) + (v_image - v_min * f_equalize) / v_max;\n"
	"    vec3 v_grey = vec3(dot(v_color, vec3(.299, .587, .114)));\n"
	"    gl_FragColor = vec4(v_color + f_desaturate * (v_grey - v_color), 1.0);\n"
	"}\n", "n_image|n_minima|n_maxima");

const TShaderInfo t_flaps_shader("render_flaps_fs", TShaderInfo::proc_Fragment,
	"uniform sampler2D n_image, n_noise;\n"
	"uniform float f_crossover_slope;\n\n"
	"\n"
	"void main()\n"
	"{\n"
	"    vec4 v_image = texture2D(n_image, gl_TexCoord[0].xy);\n"
	"    vec4 v_noise = texture2D(n_noise, gl_TexCoord[0].xy);\n"
	"    float f_alpha = ((v_noise.x - .5) * 2.0 + 1.0) * gl_TexCoord[0].z;\n"
	"    gl_FragColor = vec4(v_image.xyz, (f_alpha - .5) * f_crossover_slope + .5);\n"
	"}\n", "n_image|n_noise");

#endif //__SPECIALIZED_SHADERS_INCLUDED
