/*
								+----------------------------------+
								|                                  |
								|  ***  berLame GLUT window  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Tga.h"
#include "../berLame_src/Vector.h"
#include "../berLame_src/MinMax.h"

static Vector2f v_mouse(0, 0);
// mouse position in OpenGL coords

static int n_width = 512, n_height = 512;
// window size

static CGLState *p_state = 0;
// OpenGL state guard

static TBmp *p_display = 0;
// displayed bitmap

static TBmp *p_bitmap = 0;
static TBmp *p_min = 0;
static TBmp *p_smooth_min = 0;
static TBmp *p_max = 0;
static TBmp *p_smooth_max = 0;
static TBmp *p_equalized = 0;
// source image

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera
}

// here we render frame
void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(p_display) {
		float f_x_off = max(.0f, float(n_width - p_display->n_width) / n_width);
		float f_y_off = max(.0f, float(n_height - p_display->n_height) / n_height);
		glRasterPos2f(-1 + f_x_off, 1 - f_y_off);
		glPixelZoom(1, -1);
		glDrawPixels(p_display->n_width, p_display->n_height,
			GL_RGBA, GL_UNSIGNED_BYTE, p_display->p_buffer);
	}

	float f_pixel_x = 2.0f / n_width;
	float f_pixel_y = 2.0f / n_height;
	p_state->Color3f(1, 1, 1);
	glBegin(GL_LINES);
	glVertex2f(v_mouse.x - 2 * f_pixel_x, v_mouse.y);
	glVertex2f(v_mouse.x + 3 * f_pixel_x, v_mouse.y);
	glVertex2f(v_mouse.x, v_mouse.y - 2 * f_pixel_y);
	glVertex2f(v_mouse.x, v_mouse.y + 3 * f_pixel_y);
	glEnd();
	// draw cross under mouse cursor

	glutSwapBuffers();
	// display results
}

static int n_button;

void onMouseClick(int _n_button, int n_state, int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	n_button = _n_button;

	switch(_n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	switch(n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onPassiveMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;
	glutPostRedisplay();
}

void onKeyPressed(unsigned char n_key_code, int n_x, int n_y)
{
	if(n_key_code == 27) {
		delete p_state;
		delete[] p_bitmap->p_buffer;
		delete[] p_min->p_buffer;
		delete[] p_max->p_buffer;
		delete[] p_smooth_min->p_buffer;
		delete[] p_smooth_max->p_buffer;
		delete[] p_equalized->p_buffer;
		delete p_bitmap;
		delete p_min;
		delete p_max;
		delete p_smooth_min;
		delete p_smooth_max;
		delete p_equalized;
		exit(0);
	} else {
		switch(n_key_code) {
		case 'i':
			p_display = p_bitmap;
			break;
		case 'g':
			p_display = p_smooth_min;
			break;
		case 'r':
			p_display = p_smooth_max;
			break;
		case 'e':
			p_display = p_equalized;
			break;
		}
		glutPostRedisplay();
	}
}

TBmp *p_CloneImage(const TBmp *p_src, bool b_contents_important = false)
{
	TBmp *p_dest;
	if(!(p_dest = new TBmp))
		return 0;
	*p_dest = *p_src;
	if(!(p_dest->p_buffer = new uint32_t[p_dest->n_width * p_dest->n_height])) {
		delete p_dest;
		return 0;
	}
	// alloc a new bitmap

	if(b_contents_important) {
		memcpy(p_dest->p_buffer, p_src->p_buffer,
			p_dest->n_width * p_dest->n_height * sizeof(uint32_t));
	}
	// copy image contents if required

	return p_dest;
}

TBmp *p_Greyscale(const TBmp *p_src)
{
	if(p_src->b_grayscale) {
		TBmp *p_dest;
		if(!(p_dest = p_CloneImage(p_src, true)))
			return 0;
		// alloc dest bitmap

		return p_dest;
		// greyscale already
	} else {
		TBmp *p_dest;
		if(!(p_dest = p_CloneImage(p_src, false)))
			return 0;
		// alloc dest bitmap

		p_dest->b_grayscale = true;
		// we're now greyscale

		if(p_src->b_alpha) {
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(const uint32_t *p_src_ptr = p_src->p_buffer,
			   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
			   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
				uint32_t n_src = *p_src_ptr;
				// source pixel

				uint16_t n_red_w = uint16_t(.299f * 0x10000);
				uint16_t n_green_w = uint16_t(.587f * 0x10000);
				uint16_t n_blue_w = uint16_t(.114f * 0x10000);
				// RGB to Y weights

				uint32_t n_grey = ((((n_src >> 16) & 0xff) * n_red_w) +
								  (((n_src >> 8) & 0xff) * n_green_w) +
								  ((n_src & 0xff) * n_blue_w)) >> 16;
				// calculate grey

				*p_dest_ptr = (n_src & 0xff000000) | (n_grey << 16) | (n_grey << 8) | n_grey;
				// convert to RGB, keep alpha
			}
		} else {
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(const uint32_t *p_src_ptr = p_src->p_buffer,
			   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
			   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
				uint32_t n_src = *p_src_ptr;
				// source pixel

				uint16_t n_red_w = uint16_t(.299f * 0x10000);
				uint16_t n_green_w = uint16_t(.587f * 0x10000);
				uint16_t n_blue_w = uint16_t(.114f * 0x10000);
				// RGB to Y weights

				uint32_t n_grey = ((((n_src >> 16) & 0xff) * n_red_w) +
								  (((n_src >> 8) & 0xff) * n_green_w) +
								  ((n_src & 0xff) * n_blue_w)) >> 16;
				// calculate grey

				*p_dest_ptr = 0xff000000 | (n_grey << 16) | (n_grey << 8) | n_grey;
				// convert to RGB, alpha is 0xff
			}
		}
		// greyscale conversion

		return p_dest;
	}
}

TBmp *p_Threshold(const TBmp *p_src, int n_thresh)
{
	TBmp *p_dest;
	if(!(p_dest = p_CloneImage(p_src, false)))
		return 0;
	// alloc dest bitmap

	p_dest->b_alpha = false;
	if(p_src->b_grayscale) {
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			uint32_t n_src = *p_src_ptr;
			// source pixel

			*p_dest_ptr = 0xff000000 | (((n_src & 0xff) > n_thresh) * 0xffffff);
			// blaw threshold
		}
	} else {
		const uint32_t n_thresh2 = n_thresh << 8;
		const uint32_t n_thresh3 = n_thresh << 16;

		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			uint32_t n_src = *p_src_ptr;
			// source pixel

			*p_dest_ptr = 0xff000000 |
				(((n_src & 0xff) > n_thresh) * 0xff) |
				(((n_src & 0xff00) > n_thresh2) * 0xff00) |
				(((n_src & 0xff0000) > n_thresh3) * 0xff0000);
			// rgb threshold
		}
	}

	return p_dest;
}

float *p_CreateGauss_2D(unsigned int n_width, unsigned int n_height, float f_variance2 = .1.0f)
{
	float *p_kernel;
	if(!(p_kernel = new float[n_width * n_height]))
		return 0;
	// alloc kernel

	float f_sum = 0;
	float *p_kernel_ptr = p_kernel;
	for(unsigned int y = 0; y < n_height; ++ y) {
		float f_dy = y - n_height * .5f;
		for(unsigned int x = 0; x < n_width; ++ x, ++ p_kernel_ptr) {
			float f_dx = x - n_width * .5f;
			float f_r = float(sqrt(f_dx * f_dx + f_dy * f_dy)) / (n_width * .5f);
			// calc function argument

			if((*p_kernel_ptr = (float)exp(-.5f * f_r * f_r / f_variance2)) < 0)
				*p_kernel_ptr = 0;
			f_sum += *p_kernel_ptr;
		}
	}
	f_sum = 1 / f_sum;
	for(unsigned int i = 0, n = n_width * n_height; i < n; ++ i)
		p_kernel[i] *= f_sum;
	// calculate normalized gauss kernel

	return p_kernel;
}

float *p_CreateGauss_1D(unsigned int n_width, float f_variance2 = 1.0f)
{
	float *p_kernel;
	if(!(p_kernel = new float[n_width]))
		return 0;
	// alloc kernel

	float f_sum = 0;
	float *p_kernel_ptr = p_kernel;
	for(unsigned int x = 0; x < n_width; ++ x, ++ p_kernel_ptr) {
		float f_dx = x - n_width * .5f;
		float f_r = f_dx / (n_width * .5f);
		// calc function argument

		if((*p_kernel_ptr = (float)exp(-.5f * f_r * f_r / f_variance2)) < 0)
			*p_kernel_ptr = 0;
		f_sum += *p_kernel_ptr;
	}
	f_sum = 1 / f_sum;
	for(unsigned int i = 0; i < n_width; ++ i)
		p_kernel[i] *= f_sum;
	// calculate normalized gauss kernel

	return p_kernel;
}

bool ImageImageSub(TBmp *p_dest, const TBmp *p_src)
{
	if(p_dest->n_width != p_src->n_width || p_dest->n_height != p_src->n_height)
		return false;

	p_dest->b_alpha = false;
	// this filter doesn't count with alpha

	if(p_src->b_grayscale && p_dest->b_grayscale) {
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			int32_t n_src = *p_src_ptr & 0xff;
			int32_t n_dest = *p_dest_ptr & 0xff;
			// source, dest pixel

			if(n_dest - n_src <= 0)
				*p_dest_ptr = 0xff000000;
			else {
				int32_t n_grey = n_dest - n_src;
				*p_dest_ptr = 0xff000000 | (n_grey << 16) | (n_grey << 8) | n_grey;
			}
			// greyscale subtract
		}
	} else {
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			uint32_t n_src = *p_src_ptr;
			uint32_t n_dest = *p_dest_ptr;
			// source, dest pixel

			uint32_t n_dest2 = 0xff000000;
			int32_t n_red;
			if((n_red = (n_dest & 0xff0000) - (n_src & 0xff0000)) > 0)
				n_dest2 |= n_red;
			int32_t n_green;
			if((n_green = (n_dest & 0x00ff00) - (n_src & 0x00ff00)) > 0)
				n_dest2 |= n_green;
			int32_t n_blue;
			if((n_blue = (n_dest & 0x0000ff) - (n_src & 0x0000ff)) > 0)
				n_dest2 |= n_blue;
			*p_dest_ptr = n_dest2;
			// rgb subtract
		}
	}

	return true;
}

void ImageConstAdd(TBmp *p_dest, uint32_t n_add)
{
	p_dest->b_alpha = false;
	// this filter doesn't count with alpha

	if(p_dest->b_grayscale) {
		for(uint32_t *p_dest_ptr = p_dest->p_buffer,
		   *p_end = p_dest->p_buffer + p_dest->n_width * p_dest->n_height;
		   p_dest_ptr != p_end; ++ p_dest_ptr) {
			int32_t n_src = *p_dest_ptr & 0xff;
			// source, dest pixel

			if(n_src + n_add >= 0xff)
				*p_dest_ptr = 0xffffffff;
			else {
				int32_t n_grey = n_src + n_add;
				*p_dest_ptr = 0xff000000 | (n_grey << 16) | (n_grey << 8) | n_grey;
			}
			// greyscale add
		}
	} else {
		if(n_add > 0xff)
			n_add = 0xff;
		const uint32_t n_add1 = n_add << 8;
		const uint32_t n_add2 = n_add << 16;

		for(uint32_t *p_dest_ptr = p_dest->p_buffer,
		   *p_end = p_dest->p_buffer + p_dest->n_width * p_dest->n_height;
		   p_dest_ptr != p_end; ++ p_dest_ptr) {
			int32_t n_src = *p_dest_ptr;
			// source, dest pixel

			uint32_t n_dest2 = 0xff000000;
			uint32_t n_red;
			if((n_red = (n_src & 0xff0000) + n_add2) >= 0xff0000)
				n_dest2 |= 0xff0000;
			else
				n_dest2 |= n_red & 0xff0000;
			uint32_t n_green;
			if((n_green = (n_src & 0x00ff00) + n_add1) >= 0x00ff00)
				n_dest2 |= 0x00ff00;
			else
				n_dest2 |= n_green & 0x00ff00;
			uint32_t n_blue;
			if((n_blue = (n_src & 0x0000ff) + n_add) >= 0x0000ff)
				n_dest2 |= 0x0000ff;
			else
				n_dest2 |= n_blue & 0x0000ff;
			*p_dest_ptr = n_dest2;
			// rgb add
		}
	}
}

bool ImageImageDiv(TBmp *p_dest, const TBmp *p_src)
{
	if(p_dest->n_width != p_src->n_width || p_dest->n_height != p_src->n_height)
		return false;

	p_dest->b_alpha = false;
	// this filter doesn't count with alpha

	if(p_src->b_grayscale && p_dest->b_grayscale) {
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			int32_t n_src = *p_src_ptr & 0xff;
			int32_t n_dest = *p_dest_ptr & 0xff;
			// source, dest pixel

			int32_t n_grey;
			if(!n_src || (n_grey = (n_dest * 0xff) / n_src) >= 0xff)
				*p_dest_ptr = 0xffffffff;
			else
				*p_dest_ptr = 0xff000000 | (n_grey << 16) | (n_grey << 8) | n_grey;
			// greyscale division
		}
	} else {
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(const uint32_t *p_src_ptr = p_src->p_buffer,
		   *p_end = p_src->p_buffer + p_src->n_width * p_src->n_height;
		   p_src_ptr != p_end; ++ p_src_ptr, ++ p_dest_ptr) {
			uint32_t n_src = *p_src_ptr;
			uint32_t n_dest = *p_dest_ptr;
			// source, dest pixel

			uint32_t n_dest2 = 0xff000000;
			int32_t n_red;
			if(!(n_src & 0xff0000) || (n_red = ((n_dest & 0xff0000) * 0xff) / ((n_src >> 16) & 0xff)) >= 0xff0000)
				n_dest2 |= 0xff0000;
			else
				n_dest2 |= n_red & 0xff0000;
			int32_t n_green;
			if(!(n_src & 0x00ff00) || (n_green = ((n_dest & 0x00ff00) * 0xff) / ((n_src >> 8) & 0xff)) >= 0x00ff00)
				n_dest2 |= 0x00ff00;
			else
				n_dest2 |= n_green & 0x00ff00;
			int32_t n_blue;
			if(!(n_src & 0x0000ff) || (n_blue = ((n_dest & 0x0000ff) * 0xff) / (n_src & 0xff)) >= 0x0000ff)
				n_dest2 |= 0x0000ff;
			else
				n_dest2 |= n_blue & 0x0000ff;
			*p_dest_ptr = n_dest2;
			// rgb division
		}
	}

	return true;
}

TBmp *p_GaussFilter(const TBmp *p_src, int n_radius, bool b_separable)
{
	if(b_separable) {
		int n_diameter = n_radius * 2;
		float *p_kernel = p_CreateGauss_1D(n_diameter);
		// create gauss kernel

		TBmp *p_dest;
		if(!(p_dest = p_CloneImage(p_src, false))) {
			delete[] p_kernel;
			return 0;
		}
		// alloc dest bitmap

		TBmp *p_temp;
		if(!(p_temp = p_CloneImage(p_src, false))) {
			delete[] p_kernel;
			delete[] p_dest->p_buffer;
			delete p_dest;
			return 0;
		}
		// alloc dest bitmap

		p_dest->b_alpha = false;
		// this filter doesn't keep alpha

		if(p_src->b_grayscale) {
			const uint32_t *p_src_ptr = p_src->p_buffer;
			uint32_t *p_dest_ptr = p_temp->p_buffer;
			for(int j = 0, m = p_src->n_height; j < m; ++ j) {
				for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float f_accum = 0;

					float *p_kernel_ptr = p_kernel;
					for(int dx = 0, x = i - n_radius; dx < n_diameter; ++ dx, ++ x, ++ p_kernel_ptr) {
						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius)
							n_src = p_src_ptr[x + n * j];
						else
							n_src = p_src_ptr[min(n - 1, max(0, x)) + n * j];
						// this gets optimized out of the loop

						f_accum += (n_src & 0x0000ff) * *p_kernel_ptr;
					}
					// convolve

					if(f_accum >= 0xff)
						*p_dest_ptr = 0xffffffff;
					else if(f_accum > 0) {
						uint32_t n_clamped = uint32_t(f_accum) & 0xff;
						*p_dest_ptr = 0xff000000 | (n_clamped << 16) | (n_clamped << 8) | n_clamped;
					} else
						*p_dest_ptr = 0xff000000;
					// clamp
				}
			}
		} else {
			const uint32_t *p_src_ptr = p_src->p_buffer;
			uint32_t *p_dest_ptr = p_temp->p_buffer;
			for(int j = 0, m = p_src->n_height; j < m; ++ j) {
				for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float p_accum[3] = {0};

					float *p_kernel_ptr = p_kernel;
					for(int dx = 0, x = i - n_radius; dx < n_diameter; ++ dx, ++ x, ++ p_kernel_ptr) {
						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius)
							n_src = p_src_ptr[x + n * j];
						else
							n_src = p_src_ptr[min(n - 1, max(0, x)) + n * j];
						// this gets optimized out of the loop

						p_accum[0] += (n_src & 0x0000ff) * *p_kernel_ptr;
						p_accum[1] += (n_src & 0x00ff00) * *p_kernel_ptr;
						p_accum[2] += (n_src & 0xff0000) * *p_kernel_ptr;
					}
					// convolve

					uint32_t n_clamped = 0xff000000;
					if(p_accum[0] >= 0x0000ff)
						n_clamped |= 0x0000ff;
					else if(p_accum[0] > 0)
						n_clamped |= uint32_t(p_accum[0]) & 0x0000ff;
					if(p_accum[1] >= 0x00ff00)
						n_clamped |= 0x00ff00;
					else if(p_accum[1] > 0)
						n_clamped |= uint32_t(p_accum[1]) & 0x00ff00;
					if(p_accum[2] >= 0xff0000)
						n_clamped |= 0xff0000;
					else if(p_accum[2] > 0)
						n_clamped |= uint32_t(p_accum[2]) & 0xff0000;
					// clamp

					*p_dest_ptr = n_clamped;
				}
			}
		}

		if(p_temp->b_grayscale) {
			const uint32_t *p_src_ptr = p_temp->p_buffer;
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(int j = 0, m = p_temp->n_height; j < m; ++ j) {
				for(int i = 0, n = p_temp->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float f_accum = 0;

					float *p_kernel_ptr = p_kernel;
					for(int dy = 0, y = j - n_radius; dy < n_diameter; ++ dy, ++ y, ++ p_kernel_ptr) {
						uint32_t n_src;
						if(j >= n_radius && j < n - n_radius)
							n_src = p_src_ptr[i + n * y];
						else
							n_src = p_src_ptr[i + n * min(m - 1, max(0, y))];
						// this gets optimized out of the loop

						f_accum += (n_src & 0x0000ff) * *p_kernel_ptr;
					}
					// convolve

					if(f_accum >= 0xff)
						*p_dest_ptr = 0xffffffff;
					else if(f_accum > 0) {
						uint32_t n_clamped = uint32_t(f_accum) & 0xff;
						*p_dest_ptr = 0xff000000 | (n_clamped << 16) | (n_clamped << 8) | n_clamped;
					} else
						*p_dest_ptr = 0xff000000;
					// clamp
				}
			}
		} else {
			const uint32_t *p_src_ptr = p_temp->p_buffer;
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(int j = 0, m = p_temp->n_height; j < m; ++ j) {
				for(int i = 0, n = p_temp->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float p_accum[3] = {0};

					float *p_kernel_ptr = p_kernel;
					for(int dy = 0, y = j - n_radius; dy < n_diameter; ++ dy, ++ y, ++ p_kernel_ptr) {
						uint32_t n_src;
						if(j >= n_radius && j < n - n_radius)
							n_src = p_src_ptr[i + n * y];
						else
							n_src = p_src_ptr[i + n * min(m - 1, max(0, y))];
						// this gets optimized out of the loop

						p_accum[0] += (n_src & 0x0000ff) * *p_kernel_ptr;
						p_accum[1] += (n_src & 0x00ff00) * *p_kernel_ptr;
						p_accum[2] += (n_src & 0xff0000) * *p_kernel_ptr;
					}
					// convolve

					uint32_t n_clamped = 0xff000000;
					if(p_accum[0] >= 0x0000ff)
						n_clamped |= 0x0000ff;
					else if(p_accum[0] > 0)
						n_clamped |= uint32_t(p_accum[0]) & 0x0000ff;
					if(p_accum[1] >= 0x00ff00)
						n_clamped |= 0x00ff00;
					else if(p_accum[1] > 0)
						n_clamped |= uint32_t(p_accum[1]) & 0x00ff00;
					if(p_accum[2] >= 0xff0000)
						n_clamped |= 0xff0000;
					else if(p_accum[2] > 0)
						n_clamped |= uint32_t(p_accum[2]) & 0xff0000;
					// clamp

					*p_dest_ptr = n_clamped;
				}
			}
		}

		delete[] p_kernel;
		delete[] p_temp->p_buffer;
		delete p_temp;
		// cleanup

		return p_dest;
	} else {
		int n_diameter = n_radius * 2;
		float *p_kernel = p_CreateGauss_2D(n_diameter, n_diameter, .707f);
		// create gauss kernel

		TBmp *p_dest;
		if(!(p_dest = p_CloneImage(p_src, false))) {
			delete[] p_kernel;
			return 0;
		}
		// alloc dest bitmap

		p_dest->b_alpha = false;
		// this filter doesn't keep alpha

		if(p_src->b_grayscale) {
			const uint32_t *p_src_ptr = p_src->p_buffer;
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(int j = 0, m = p_src->n_height; j < m; ++ j) {
				for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float f_accum = 0;

					float *p_kernel_ptr = p_kernel;
					for(int dy = 0, y = j - n_radius; dy < n_diameter; ++ dy, ++ y) {
						for(int dx = 0, x = i - n_radius; dx < n_diameter; ++ dx, ++ x, ++ p_kernel_ptr) {
							uint32_t n_src;
							if(i >= n_radius && i < n - n_radius) {
								if(j >= n_radius && j < m - n_radius)
									n_src = p_src_ptr[x + n * y];
								else
									n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
							} else {
								if(j >= n_radius && j < m - n_radius)
									n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
								else
									n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
							}
							// this gets optimized out of the loop

							f_accum += (n_src & 0x0000ff) * *p_kernel_ptr;
						}
					}
					// convolve

					if(f_accum >= 0xff)
						*p_dest_ptr = 0xffffffff;
					else if(f_accum > 0) {
						uint32_t n_clamped = uint32_t(f_accum) & 0xff;
						*p_dest_ptr = 0xff000000 | (n_clamped << 16) | (n_clamped << 8) | n_clamped;
					} else
						*p_dest_ptr = 0xff000000;
					// clamp
				}
			}
		} else {
			const uint32_t *p_src_ptr = p_src->p_buffer;
			uint32_t *p_dest_ptr = p_dest->p_buffer;
			for(int j = 0, m = p_src->n_height; j < m; ++ j) {
				for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
					float p_accum[3] = {0};

					float *p_kernel_ptr = p_kernel;
					for(int dy = 0, y = j - n_radius; dy < n_diameter; ++ dy, ++ y) {
						for(int dx = 0, x = i - n_radius; dx < n_diameter; ++ dx, ++ x, ++ p_kernel_ptr) {
							uint32_t n_src;
							if(i >= n_radius && i < n - n_radius) {
								if(j >= n_radius && j < m - n_radius)
									n_src = p_src_ptr[x + n * y];
								else
									n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
							} else {
								if(j >= n_radius && j < m - n_radius)
									n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
								else
									n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
							}
							// this gets optimized out of the loop

							p_accum[0] += (n_src & 0x0000ff) * *p_kernel_ptr;
							p_accum[1] += (n_src & 0x00ff00) * *p_kernel_ptr;
							p_accum[2] += (n_src & 0xff0000) * *p_kernel_ptr;
						}
					}
					// convolve

					uint32_t n_clamped = 0xff000000;
					if(p_accum[0] >= 0x0000ff)
						n_clamped |= 0x0000ff;
					else if(p_accum[0] > 0)
						n_clamped |= uint32_t(p_accum[0]) & 0x0000ff;
					if(p_accum[1] >= 0x00ff00)
						n_clamped |= 0x00ff00;
					else if(p_accum[1] > 0)
						n_clamped |= uint32_t(p_accum[1]) & 0x00ff00;
					if(p_accum[2] >= 0xff0000)
						n_clamped |= 0xff0000;
					else if(p_accum[2] > 0)
						n_clamped |= uint32_t(p_accum[2]) & 0xff0000;
					// clamp

					*p_dest_ptr = n_clamped;
				}
			}
		}

		delete[] p_kernel;

		return p_dest;
	}
}

TBmp *p_LocalMin(const TBmp *p_src, int n_radius)
{
	TBmp *p_dest;
	if(!(p_dest = p_CloneImage(p_src, false)))
		return 0;
	// alloc dest bitmap

	int n_side = n_radius * 2 + 1;
	bool *p_mask;
	if(!(p_mask = new bool[n_side * n_side])) {
		delete[] p_dest->p_buffer;
		delete p_dest;
		return 0;
	}
	for(int j = 0; j < n_side; ++ j) {
		for(int i = 0; i < n_side; ++ i) {
			int dx = i - n_radius;
			int dy = j - n_radius;
			p_mask[i + n_side * j] = sqrt(dx * dx + dy * dy) <= n_radius;
		}
	}
	// create circular mask

	p_dest->b_alpha = false;
	if(p_src->b_grayscale) {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height; j < m; ++ j) {
			for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
				uint32_t n_min_grey = 0xff;

				bool *p_mask_ptr = p_mask;
				for(int dy = 0, y = j - n_radius; dy < n_side; ++ dy, ++ y) {
					for(int dx = 0, x = i - n_radius; dx < n_side; ++ dx, ++ x, ++ p_mask_ptr) {
						if(!*p_mask_ptr)
							continue;

						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius) {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[x + n * y];
							else
								n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
						} else {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
							else
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
						}
						// this gets optimized out of the loop

						uint32_t n_grey = n_src & 0x0000ff;
						if(n_min_grey > n_grey)
							n_min_grey = n_grey;
					}
				}

				*p_dest_ptr = 0xff000000 | (n_min_grey << 16) | (n_min_grey << 8) | n_min_grey;
			}
		}
	} else {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height; j < m; ++ j) {
			for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
				uint32_t n_min_r = 0xff0000, n_min_g = 0xff00, n_min_b = 0xff;

				bool *p_mask_ptr = p_mask;
				for(int dy = 0, y = j - n_radius; dy < n_side; ++ dy, ++ y) {
					for(int dx = 0, x = i - n_radius; dx < n_side; ++ dx, ++ x, ++ p_mask_ptr) {
						if(!*p_mask_ptr)
							continue;

						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius) {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[x + n * y];
							else
								n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
						} else {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
							else
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
						}
						// this gets optimized out of the loop

						uint32_t n_r = n_src & 0xff0000;
						if(n_min_r > n_r)
							n_min_r = n_r;
						uint32_t n_g = n_src & 0x00ff00;
						if(n_min_g > n_g)
							n_min_g = n_g;
						uint32_t n_b = n_src & 0x0000ff;
						if(n_min_b > n_b)
							n_min_b = n_b;
					}
				}

				*p_dest_ptr = 0xff000000 | n_min_r | n_min_g | n_min_b;
			}
		}
	}

	delete[] p_mask;

	return p_dest;
}

TBmp *p_LocalMax(const TBmp *p_src, int n_radius)
{
	TBmp *p_dest;
	if(!(p_dest = p_CloneImage(p_src, false)))
		return 0;
	// alloc dest bitmap

	int n_side = n_radius * 2 + 1;
	bool *p_mask;
	if(!(p_mask = new bool[n_side * n_side])) {
		delete[] p_dest->p_buffer;
		delete p_dest;
		return 0;
	}
	for(int j = 0; j < n_side; ++ j) {
		for(int i = 0; i < n_side; ++ i) {
			int dx = i - n_radius;
			int dy = j - n_radius;
			p_mask[i + n_side * j] = sqrt(dx * dx + dy * dy) <= n_radius;
		}
	}
	// create circular mask

	p_dest->b_alpha = false;
	if(p_src->b_grayscale) {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height; j < m; ++ j) {
			for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
				uint32_t n_max_grey = 0;

				bool *p_mask_ptr = p_mask;
				for(int dy = 0, y = j - n_radius; dy < n_side; ++ dy, ++ y) {
					for(int dx = 0, x = i - n_radius; dx < n_side; ++ dx, ++ x, ++ p_mask_ptr) {
						if(!*p_mask_ptr)
							continue;

						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius) {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[x + n * y];
							else
								n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
						} else {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
							else
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
						}
						// this gets optimized out of the loop

						uint32_t n_grey = n_src & 0x0000ff;
						if(n_max_grey < n_grey)
							n_max_grey = n_grey;
					}
				}

				*p_dest_ptr = 0xff000000 | (n_max_grey << 16) | (n_max_grey << 8) | n_max_grey;
			}
		}
	} else {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height; j < m; ++ j) {
			for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr) {
				uint32_t n_max_r = 0, n_max_g = 0, n_max_b = 0;

				bool *p_mask_ptr = p_mask;
				for(int dy = 0, y = j - n_radius; dy < n_side; ++ dy, ++ y) {
					for(int dx = 0, x = i - n_radius; dx < n_side; ++ dx, ++ x, ++ p_mask_ptr) {
						if(!*p_mask_ptr)
							continue;

						uint32_t n_src;
						if(i >= n_radius && i < n - n_radius) {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[x + n * y];
							else
								n_src = p_src_ptr[x + n * min(m - 1, max(0, y))];
						} else {
							if(j >= n_radius && j < m - n_radius)
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * y];
							else
								n_src = p_src_ptr[min(n - 1, max(0, x)) + n * min(m - 1, max(0, y))];
						}
						// this gets optimized out of the loop

						uint32_t n_r = n_src & 0xff0000;
						if(n_max_r < n_r)
							n_max_r = n_r;
						uint32_t n_g = n_src & 0x00ff00;
						if(n_max_g < n_g)
							n_max_g = n_g;
						uint32_t n_b = n_src & 0x0000ff;
						if(n_max_b < n_b)
							n_max_b = n_b;
					}
				}

				*p_dest_ptr = 0xff000000 | n_max_r | n_max_g | n_max_b;
			}
		}
	}

	delete[] p_mask;

	return p_dest;
}

TBmp *p_RobertCross(const TBmp *p_src)
{
	TBmp *p_dest;
	if(!(p_dest = p_CloneImage(p_src, false)))
		return 0;
	// alloc dest bitmap

	p_dest->b_alpha = false;
	if(p_src->b_grayscale) {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height - 1; j < m; ++ j, ++ p_src_ptr, ++ p_dest_ptr) {
			for(int i = 0, n = p_src->n_width - 1, o = p_src->n_width + 1, w = p_src->n_width;
			   i < n; ++ i, ++ p_src_ptr, ++ p_dest_ptr) {
				int n_a = (*p_src_ptr & 0xff) - (p_src_ptr[o] & 0xff);
				int n_b = (p_src_ptr[1] & 0xff) - (p_src_ptr[w] & 0xff);
				int n_dest = int(sqrt(n_a * n_a + n_b * n_b));
				if(n_dest > 0xff)
					n_dest = 0xff;
				else if(n_dest < 0)
					n_dest = 0;
				*p_dest_ptr = 0xff000000 | (n_dest << 16) | (n_dest << 8) | n_dest;
			}
			*p_dest_ptr = 0xff000000; // last pixel of every scanline is black
		}
		// apply robert-cross operator on inner image

		for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr)
			*p_dest_ptr = 0xff000000;
		// last scanline is black
	} else {
		const uint32_t *p_src_ptr = p_src->p_buffer;
		uint32_t *p_dest_ptr = p_dest->p_buffer;
		for(int j = 0, m = p_src->n_height - 1; j < m; ++ j, ++ p_src_ptr, ++ p_dest_ptr) {
			for(int i = 0, n = p_src->n_width - 1, o = p_src->n_width + 1, w = p_src->n_width;
			   i < n; ++ i, ++ p_src_ptr, ++ p_dest_ptr) {
				uint32_t n_pixel00 = *p_src_ptr;
				uint32_t n_pixel10 = p_src_ptr[1];
				uint32_t n_pixel01 = p_src_ptr[w];
				uint32_t n_pixel11 = p_src_ptr[o];
				int n_r_a = (n_pixel00 & 0xff0000) - (n_pixel11 & 0xff0000);
				int n_r_b = (n_pixel10 & 0xff0000) - (n_pixel01 & 0xff0000);
				int n_r_dest = int(sqrt(n_r_a * n_r_a + n_r_b * n_r_b));
				int n_g_a = (n_pixel00 & 0xff00) - (n_pixel11 & 0xff00);
				int n_g_b = (n_pixel10 & 0xff00) - (n_pixel01 & 0xff00);
				int n_g_dest = int(sqrt(n_g_a * n_g_a + n_g_b * n_g_b));
				int n_b_a = (n_pixel00 & 0xff) - (n_pixel11 & 0xff);
				int n_b_b = (n_pixel10 & 0xff) - (n_pixel01 & 0xff);
				int n_b_dest = int(sqrt(n_b_a * n_b_a + n_b_b * n_b_b));
				if(n_r_dest > 0xff0000)
					n_r_dest = 0xff0000;
				else if(n_r_dest < 0x010000)
					n_r_dest = 0;
				else
					n_r_dest &= 0xff00;
				if(n_g_dest > 0xff00)
					n_g_dest = 0xff00;
				else if(n_g_dest < 0x0100)
					n_g_dest = 0;
				else
					n_g_dest &= 0xff00;
				if(n_b_dest > 0xff)
					n_b_dest = 0xff;
				else if(n_b_dest < 0x01)
					n_b_dest = 0;
				else
					n_b_dest &= 0xff;
				*p_dest_ptr = 0xff000000 | n_r_dest | n_g_dest | n_b_dest;
			}
			*p_dest_ptr = 0xff000000; // last pixel of every scanline is black
		}
		// apply robert-cross operator on inner image

		for(int i = 0, n = p_src->n_width; i < n; ++ i, ++ p_dest_ptr)
			*p_dest_ptr = 0xff000000;
		// last scanline is black
	}

	return p_dest;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("berLame GLUT window");
	// init OpenGL using GLUT

	if(!(p_state = new CGLState(/*wglGetCurrentDC(), wglGetCurrentContext()*/))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create OpenGL state guard

	if(!(p_bitmap = CTgaCodec::p_Load_TGA("test_bark.tga"))) {
		fprintf(stderr, "error: failed to load test.tga\n");
		return -1;
	}
	if(!(p_min = p_LocalMin(p_bitmap, 16)) ||
	   !(p_smooth_min = p_GaussFilter(p_min, 32, true)) ||
	   !(p_max = p_LocalMax(p_bitmap, 16)) ||
	   !(p_smooth_max = p_GaussFilter(p_max, 32, true)) ||
	   !(p_equalized = p_CloneImage(p_bitmap, true))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// now we could load textures and or do other initialization

	ImageImageSub(p_equalized, p_smooth_min);
	ImageImageSub(p_smooth_max, p_smooth_min);
	ImageConstAdd(p_smooth_max, 1); // ad headroom
	ImageImageDiv(p_equalized, p_smooth_max);
	// per-pixel histogram equalization

	p_display = p_bitmap;
	printf("press: i = src image, g = greyscale, r = robert-cross applied, e = edges\n");

	// todo - morphologic edge post-processing?
	// todo - edge vectorization

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	glutKeyboardFunc(onKeyPressed);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
