/*
								+----------------------------------+
								|                                  |
								|  ***  berLame GLUT window  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/TextureUtil.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/Vector.h"
#include "../berLame_src/MinMax.h"
#include "ImageProc.h"
#include "SpecialShaders.h"

static Vector2f v_mouse(0, 0);
// mouse position in OpenGL coords

static float f_scene_z = -200;
static float f_angle_y = 0, f_angle_x = 0;
// scene rotation

static Vector2f v_scene_pos(0, 0);
// scene position

static int n_width = 1024, n_height = 768;
// window size

static CGLState *p_state = 0;
// OpenGL state guard

static CGLTexture_2D *p_src_image = 0;
// source image

static CGLTexture_2D *p_temp_tex[4] = {0};
static CGLTexture_2D *p_min = 0;
static CGLTexture_2D *p_min_smooth = 0;
static CGLTexture_2D *p_max = 0;
static CGLTexture_2D *p_max_smooth = 0;
static CGLTexture_2D *p_equalized = 0;
static CGLTexture_2D *p_noise = 0;
static CGLTexture_2D *p_repeatable = 0;
// processed images

static CGLFrameBuffer *p_fbo = 0;
// framebuffer objects

static CGLShaderObject *p_perpix_equalize = 0, *p_perpix_grey_equalize = 0;
static CGLShaderObject *p_flaps_shader = 0;
// special shaders

static CGLImageProc image_proc_manager; // gotta have one of those
static CGLPerlinNoise perlin_noise;
static CGLGaussFilter gauss_filter;
static CGLMinMaxFilter minmax_filter;
// common filters

enum {
	flatten_Color,
	flatten_Grey,
	flatten_Disable
};
// flatten mode

static int n_min_max_size = 15;
static int n_gauss_size = 15;
static float f_headroom = .3f;
static float f_brightness = .1f;
static float f_desaturate = 0;
static float f_equalize = 1;
static float f_noise_size = 10, f_noise_amp = .5f, f_noise_persitance = .5f;
static int n_flatten_processing = flatten_Color;
static int n_noise_octaves = 5;
static float f_seam_coverup_size = .3f, f_crossover_slope = 5;
// filter parameters

enum {
	filter_MinMax,
	filter_Gauss,
	filter_Equalizer,
	filter_Noise,
	filter_Seams
};
// filter names

static int n_adjusting = filter_Seams;
// filter we're currently adjusting

enum {
	display_Controls = 0,
	display_OutputFlybye
};
// display mode names

static int n_mode = display_Controls;
// current display mode

bool CheckOpenGL_Extensions();
CGLTexture_2D *p_CloneInputImageFmt();
bool Render_MinMax_Textures(int n_kernel_size);
bool Render_SmoothMinMax_Textures(int n_radius);
bool Render_Equalized_Texture();
bool Render_Noise_Texture(float f_scale,
	float f_amplitude, int n_octave_num, float f_persistence);
bool Render_Repeatable_Texture(float f_seam_cover_size, float f_crossover_slope);
void Cleanup();
// globals

void glutRenderString(Vector2f v_pos, const char *p_s_string, void *p_font = GLUT_BITMAP_8_BY_13)
{
	glRasterPos2f(v_pos.x, v_pos.y);
	for(; *p_s_string; ++ p_s_string)
    	glutBitmapCharacter(p_font, *p_s_string);
}

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height); // set new viewport
}

// here we render frame
void onDisplay()
{
	glViewport(0, 0, n_width, n_height);
	p_state->ClearColor4f(.85f, .85f, .85f, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	p_state->DisableDepthTest();
	p_state->DisableCullFace();
	p_state->BindProgramObject_GL2(0);
	p_state->DisableFragmentProgram();
	p_state->DisableVertexProgram();
	p_state->Color3f(1, 1, 1);
	// disable everything disturbing

	if(n_mode == display_Controls) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(-1, -1, 0);
		glScalef(2.0f / n_width, 2.0f / n_height, 1);
		// scale opengl coords to match raster coords

		CGLTexture_2D *p_show_texs[4] = {p_src_image, 0, p_equalized, 0};

		if(n_adjusting == filter_MinMax) {
			p_show_texs[1] = p_min;
			p_show_texs[3] = p_max;
		} else if(n_adjusting == filter_Gauss) {
			p_show_texs[1] = p_min_smooth;
			p_show_texs[3] = p_max_smooth;
		} else if(n_adjusting == filter_Noise) {
			p_show_texs[1] = p_noise;
			p_show_texs[3] = p_repeatable;
		} else //if(n_adjusting == filter_Seams)
			p_show_texs[3] = p_repeatable;
		// show textures dependent on which filter is being adjusted
		
		for(int i = 0; i < 4; ++ i) {
			if(!p_show_texs[i])
				continue;
			p_show_texs[i]->Bind_Enable(p_state);

			int n_space_h = n_width / 2;
			int n_space_v = n_height / 2;
			int n_org_x = n_space_h * (i % 2);
			int n_org_y = n_space_v * (1 - i / 2);
			int n_display_w = p_show_texs[i]->n_Width();
			int n_display_h = p_show_texs[i]->n_Height();
			const int n_max_w = n_width / 2 - 48;
			const int n_max_h = n_height / 2 - 48;
			if(n_display_w > n_max_w) {
				n_display_h = (n_display_h * n_max_w) / n_display_w;
				n_display_w = n_max_w;
			}
			if(n_display_h > n_max_h) {
				n_display_w = (n_display_w * n_max_h) / n_display_h;
				n_display_h = n_max_h;
			}
			int n_off_x = n_org_x + (n_space_h - n_display_w) / 2;
			int n_off_y = n_org_y + (n_space_v - n_display_h) / 2;

			glBegin(GL_QUADS);
			glTexCoord2f(0, 1);
			glVertex2i(n_off_x, n_off_y);
			glTexCoord2f(0, 0);
			glVertex2i(n_off_x, n_off_y + n_display_h);
			glTexCoord2f(1, 0);
			glVertex2i(n_off_x + n_display_w, n_off_y + n_display_h);
			glTexCoord2f(1, 1);
			glVertex2i(n_off_x + n_display_w, n_off_y);
			glEnd();
		}
		// draw images

		p_state->DisableTexture2D();
		p_state->Color3f(0, 0, 0);

		char p_s_buffer[512];
		sprintf(p_s_buffer, "min-max filter radius: %d [s/x] |"
			" gauss filter radius: %d [d/c] | headroom: %.2f [f/v] | brightness: %.2f [g/b] | equalize: %.2f [z/y] | desaturate: %.2f [1/2]",
			n_min_max_size, n_gauss_size, f_headroom, f_brightness, f_equalize, f_desaturate);
		glutRenderString(Vector2f(10, 10), p_s_buffer);
		sprintf(p_s_buffer, "noise: size %.2f [q/w], amp %.2f [e/r], octaves %d [u/i],"
			" persist %.2f [o/p] | seam size: %.1f %% [h/n] | crossover: %.2f [j/m]",
			f_noise_size, f_noise_amp, n_noise_octaves, f_noise_persitance,
			f_seam_coverup_size * 100, f_crossover_slope);
		glutRenderString(Vector2f(10, n_height - 15), p_s_buffer);
		const char *p_flatten_name_list[] = {
			"enhance colors & contrast",
			"enhance monochromatic contrast",
			"disabled"
		};
		sprintf(p_s_buffer, "flatten image: %s [a] | output size: %d x %d "
			"[press k to save, press space to view in flybye mode]",
			p_flatten_name_list[n_flatten_processing], p_repeatable->n_Width(), p_repeatable->n_Height());
		glutRenderString(Vector2f(10, n_height / 2 - 5), p_s_buffer);
	} else {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(90, (float)n_width / n_height, .1f, 10000);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glTranslatef(0, 0, f_scene_z);
		glRotatef(f_angle_x, 1, 0, 0);
		glRotatef(f_angle_y, 0, 1, 0);
		glTranslatef(v_scene_pos.x, v_scene_pos.y, 0);
		// rotate/translate scene mouse-wisely

		p_repeatable->Bind_Enable(p_state);
		p_state->Color3f(1, 1, 1);

		const float f_size_y = 10000;
		const float f_size_x = f_size_y * float(p_repeatable->n_Width()) / p_repeatable->n_Height();
		const float f_tex = 101; // must be odd
		glBegin(GL_QUADS);
		glTexCoord2f(0, f_tex);
		glVertex2i(-f_size_x, -f_size_y);
		glTexCoord2f(0, 0);
		glVertex2i(-f_size_x, f_size_y);
		glTexCoord2f(f_tex, 0);
		glVertex2i(f_size_x, f_size_y);
		glTexCoord2f(f_tex, f_tex);
		glVertex2i(f_size_x, -f_size_y);
		glEnd();

		p_state->DisableTexture2D();
		p_state->Color3f(1, 0, 0);

		glBegin(GL_LINE_LOOP);
		glVertex2i(-f_size_x / f_tex, -f_size_y / f_tex);
		glVertex2i(-f_size_x / f_tex, f_size_y / f_tex);
		glVertex2i(f_size_x / f_tex, f_size_y / f_tex);
		glVertex2i(f_size_x / f_tex, -f_size_y / f_tex);
		glEnd();
	}

	glutSwapBuffers();
	// display results
}

static int n_button;

void onMouseClick(int _n_button, int n_state, int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	n_button = _n_button;

	switch(_n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onMouseMotion(int n_x, int n_y)
{
	Vector2f v_new_mouse((2.0f * n_x) / n_width - 1, 1 - (2.0f * n_y) / n_height);
	// get current mouse position

	if(n_button == GLUT_LEFT_BUTTON) {
		f_angle_y = max(-89.0f, min(89.0f, f_angle_y + (v_new_mouse.x - v_mouse.x) * 90));
		f_angle_x = max(-89.0f, min(89.0f, f_angle_x + (v_new_mouse.y - v_mouse.y) * 90));
	} else if(n_button == GLUT_RIGHT_BUTTON) {
		f_scene_z = min(-.2f, f_scene_z +
			(v_new_mouse.x - v_mouse.x + v_new_mouse.y - v_mouse.y) * 100);
	} else
		v_scene_pos += (v_new_mouse - v_mouse) * Vector2f(float(n_width) / n_height, 1) * -f_scene_z;

	v_mouse = v_new_mouse;
	// set current mouse position

	glutPostRedisplay();
	// asume we changed something and now we want to redraw ...
}

void onPassiveMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;
	glutPostRedisplay();
}

void SaveImage()
{
	if(!p_repeatable) {
		fprintf(stderr, "error: output image doesn't exist\n");
		return;
	}

	char p_s_filename[260] = {0};
	OPENFILENAME t_open_file;
	memset(&t_open_file, 0, sizeof(OPENFILENAME));
	t_open_file.lStructSize = sizeof(OPENFILENAME);
	t_open_file.hwndOwner = NULL;
	t_open_file.lpstrFile = p_s_filename;
	t_open_file.nMaxFile = sizeof(p_s_filename);
	t_open_file.lpstrFilter = "Targa images\0*.tga\0";
	t_open_file.nFilterIndex = 1;
	t_open_file.lpstrDefExt = ".tga";
	t_open_file.lpstrFileTitle = NULL;
	t_open_file.nMaxFileTitle = 0;
	t_open_file.lpstrInitialDir = NULL;
	t_open_file.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
	t_open_file.lpstrTitle = "Save texture ...";

	if(!GetSaveFileName(&t_open_file))
		return;

	TBmp t_bmp;
	t_bmp.b_alpha = true;
	t_bmp.b_grayscale = false;
	t_bmp.n_former_bpp = 8;
	t_bmp.n_width = p_repeatable->n_Width();
	t_bmp.n_height = p_repeatable->n_Height();
	if(!(t_bmp.p_buffer = new uint32_t[t_bmp.n_width * t_bmp.n_height])) {
		fprintf(stderr, "error: not enough memory\n");
		return;
	}

	p_repeatable->Bind_Enable(p_state);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_BGRA, GL_UNSIGNED_BYTE, t_bmp.p_buffer);

	if(!CTgaCodec::Save_TGA(p_s_filename, t_bmp)) {
		delete[] t_bmp.p_buffer;
		fprintf(stderr, "error: failed to save output image\n");
		return;
	}
	delete[] t_bmp.p_buffer;

	printf("output saved to \'%s\'\n", p_s_filename);
}

void onKeyPressed(unsigned char n_key_code, int n_x, int n_y)
{
	if(n_key_code == 27) {
		Cleanup();
		exit(0);
	} else {
		n_adjusting = filter_Seams;
		switch(n_key_code) {
		case ' ':
			if(++ n_mode > display_OutputFlybye)
				n_mode = 0;
			break;
		case 'a':
			n_adjusting = filter_Equalizer;
			++ n_flatten_processing;
			if(n_flatten_processing > flatten_Disable)
				n_flatten_processing = flatten_Color;
			break;
		case 's':
			n_adjusting = filter_MinMax;
			if(n_min_max_size > 1)
				n_min_max_size -= 2;
			break;
		case 'x':
			n_adjusting = filter_MinMax;
			if(n_min_max_size < 255)
				n_min_max_size += 2;
			break;
		case 'd':
			n_adjusting = filter_Gauss;
			if(n_gauss_size > 1)
				-- n_gauss_size;
			break;
		case 'c':
			n_adjusting = filter_Gauss;
			if(n_gauss_size < 255)
				++ n_gauss_size;
			break;
		case 'f':
			n_adjusting = filter_Equalizer;
			if(f_headroom > 0)
				f_headroom = max(.0f, f_headroom - .01f);
			break;
		case 'v':
			n_adjusting = filter_Equalizer;
			if(f_headroom < 1)
				f_headroom = min(1.0f, f_headroom + .01f);
			break;
		case 'g':
			n_adjusting = filter_Equalizer;
			if(f_brightness > 0)
				f_brightness = max(.0f, f_brightness - .01f);
			break;
		case 'b':
			n_adjusting = filter_Equalizer;
			if(f_brightness < 1)
				f_brightness = min(1.0f, f_brightness + .01f);
			break;
		case '1':
			n_adjusting = filter_Equalizer;
			if(f_desaturate > 0)
				f_desaturate = max(.0f, f_desaturate - .01f);
			break;
		case '2':
			n_adjusting = filter_Equalizer;
			if(f_desaturate < 1)
				f_desaturate = min(1.0f, f_desaturate + .01f);
			break;
		case 'z':
			n_adjusting = filter_Equalizer;
			if(f_equalize > 0)
				f_equalize = max(.0f, f_equalize - .01f);
			break;
		case 'y':
			n_adjusting = filter_Equalizer;
			if(f_equalize < 10)
				f_equalize = min(10.0f, f_equalize + .01f);
			break;
		case 'q':
			n_adjusting = filter_Noise;
			if(f_noise_size > 0)
				f_noise_size = max(.0f, f_noise_size - .1f);
			break;
		case 'w':
			n_adjusting = filter_Noise;
			f_noise_size += .1f;
			break;
		case 'e':
			n_adjusting = filter_Noise;
			if(f_noise_amp > 0)
				f_noise_amp = max(.0f, f_noise_amp - .01f);
			break;
		case 'r':
			n_adjusting = filter_Noise;
			f_noise_amp += .01f;
			break;
		case 'u':
			n_adjusting = filter_Noise;
			if(n_noise_octaves > 0)
				-- n_noise_octaves;
			break;
		case 'i':
			n_adjusting = filter_Noise;
			++ n_noise_octaves;
			break;
		case 'o':
			n_adjusting = filter_Noise;
			if(f_noise_persitance > 0)
				f_noise_persitance = max(.0f, f_noise_persitance - .01f);
			break;
		case 'p':
			n_adjusting = filter_Noise;
			if(f_noise_persitance < 1)
				f_noise_persitance = min(1.0f, f_noise_persitance + .01f);
			break;
		case 'h':
			n_adjusting = filter_Seams;
			if(f_seam_coverup_size > 0)
				f_seam_coverup_size = max(.0f, f_seam_coverup_size - .01f);
			break;
		case 'n':
			n_adjusting = filter_Seams;
			if(f_seam_coverup_size < 1)
				f_seam_coverup_size = min(1.0f, f_seam_coverup_size + .01f);
			break;
		case 'j':
			n_adjusting = filter_Seams;
			if(f_crossover_slope > 1)
				f_crossover_slope = max(1.0f, f_crossover_slope - .1f);
			break;
		case 'm':
			n_adjusting = filter_Seams;
			f_crossover_slope = f_crossover_slope + .1f;
			break;
		case 'k':
			SaveImage();
			break;
		case 'l':
			if(!Render_MinMax_Textures(n_min_max_size)) {
				fprintf(stderr, "error: Render_MinMax_Textures()\n");
				Cleanup();
				exit(-1);
			}
			if(!Render_SmoothMinMax_Textures(n_gauss_size)) {
				fprintf(stderr, "error: Render_SmoothMinMax_Textures()\n");
				Cleanup();
				exit(-1);
			}
			if(!Render_Equalized_Texture()) {
				fprintf(stderr, "error: Render_Equalized_Texture()\n");
				Cleanup();
				exit(-1);
			}
			if(!Render_Noise_Texture(f_noise_size, f_noise_amp,
			   n_noise_octaves, f_noise_persitance)) {
				fprintf(stderr, "error: Render_Noise_Texture()\n");
				Cleanup();
				exit(-1);
			}
			if(!Render_Repeatable_Texture(f_seam_coverup_size, f_crossover_slope)) {
				fprintf(stderr, "error: Render_Repeatable_Texture()\n");
				Cleanup();
				exit(-1);
			}
			break;
		}
		if(p_src_image->n_Width() * p_src_image->n_Height() <= 768 * 768) {
			switch(n_adjusting) {
			case filter_MinMax:
				if(!Render_MinMax_Textures(n_min_max_size)) {
					fprintf(stderr, "error: Render_MinMax_Textures()\n");
					Cleanup();
					exit(-1);
				}
				// this case intentionally falls trough
			case filter_Gauss:
				if(!Render_SmoothMinMax_Textures(n_gauss_size)) {
					fprintf(stderr, "error: Render_SmoothMinMax_Textures()\n");
					Cleanup();
					exit(-1);
				}
				// this case intentionally falls trough
			case filter_Equalizer:
				if(!Render_Equalized_Texture()) {
					fprintf(stderr, "error: Render_Equalized_Texture()\n");
					Cleanup();
					exit(-1);
				}
				if(!Render_Repeatable_Texture(f_seam_coverup_size, f_crossover_slope)) {
					fprintf(stderr, "error: Render_Repeatable_Texture()\n");
					Cleanup();
					exit(-1);
				}
				break;
			case filter_Noise:
				if(!Render_Noise_Texture(f_noise_size, f_noise_amp,
				   n_noise_octaves, f_noise_persitance)) {
					fprintf(stderr, "error: Render_Noise_Texture()\n");
					Cleanup();
					exit(-1);
				}
				// this case intentionally falls trough
			case filter_Seams:
				if(!Render_Repeatable_Texture(f_seam_coverup_size, f_crossover_slope)) {
					fprintf(stderr, "error: Render_Repeatable_Texture()\n");
					Cleanup();
					exit(-1);
				}
				break;
			};
		}
		glutPostRedisplay();
	}
}

bool Render_MinMax_Textures(int n_kernel_size)
{
	return minmax_filter.Filter(p_state, p_min, p_max, p_temp_tex, p_src_image, n_kernel_size);
}

bool Render_SmoothMinMax_Textures(int n_radius)
{
	if(!gauss_filter.Filter(p_state, p_min_smooth, p_temp_tex[0], p_min, n_radius) ||
	   !gauss_filter.Filter(p_state, p_max_smooth, p_temp_tex[0], p_max, n_radius))
		return false;
	return true;
}

bool Render_Equalized_Texture()
{
	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!p_fbo->Bind() ||
	   !p_fbo->BindColorTexture_2D(*p_equalized, GL_TEXTURE_2D, 0))
		return false;
	// bind FBO & textures

	if(glGetError() != GL_NO_ERROR)
		return false;

	glViewport(0, 0, p_fbo->n_Width(), p_fbo->n_Height());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT);
	// viewport, matrices & clear

	p_state->ActiveTextureUnit(0);
	p_src_image->Bind_Enable(p_state);

	if(n_flatten_processing == flatten_Disable) {
		p_state->ActiveTextureUnit(1);
		p_state->DisableTexture2D();
		p_state->ActiveTextureUnit(2);
		p_state->DisableTexture2D();
		p_state->BindProgramObject(0);
		p_state->DisableVertexProgram();
		p_state->DisableFragmentProgram();
		p_state->Color3f(1, 1, 1);
		// disable other textures & shaders, just copy
	} else {
		p_state->ActiveTextureUnit(1);
		p_min_smooth->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(2);
		p_max_smooth->Bind_Enable(p_state);
		// bind outputs from the first pass

		if(n_flatten_processing == flatten_Color) {
			p_perpix_equalize->t_Find_Parameter("f_headroom").SetValue(f_headroom);
			p_perpix_equalize->t_Find_Parameter("f_brightness").SetValue(f_brightness);
			p_perpix_equalize->t_Find_Parameter("f_equalize").SetValue(f_equalize);
			p_perpix_equalize->t_Find_Parameter("f_desaturate").SetValue(f_desaturate);
			// set shader params

			if(!p_perpix_equalize->Bind(p_state))
				return false;
			// bind shader
		} else {
			p_perpix_grey_equalize->t_Find_Parameter("f_headroom").SetValue(f_headroom);
			p_perpix_grey_equalize->t_Find_Parameter("f_brightness").SetValue(f_brightness);
			p_perpix_grey_equalize->t_Find_Parameter("f_equalize").SetValue(f_equalize);
			p_perpix_grey_equalize->t_Find_Parameter("f_desaturate").SetValue(f_desaturate);
			// set shader params

			if(!p_perpix_grey_equalize->Bind(p_state))
				return false;
			// bind shader
		}
	}

	if(glGetError() != GL_NO_ERROR)
		return false;

	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(-1, -1);
	glTexCoord2f(0, 1);
	glVertex2f(-1, 1);
	glTexCoord2f(1, 1);
	glVertex2f(1, 1);
	glTexCoord2f(1, 0);
	glVertex2f(1, -1);
	glEnd();
	// draw fullscreen quad

	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!p_fbo->BindColorTexture_2D(0, GL_TEXTURE_2D, 0) ||
	   !p_fbo->Release())
		return false;
	// release textures & FBO

	for(int i = 1; i < 3; ++ i) {
		p_state->ActiveTextureUnit(i);
		p_state->DisableTexture2D();
	}
	p_state->ActiveTextureUnit(0);

	return glGetError() == GL_NO_ERROR;
}

bool Render_Noise_Texture(float f_scale, float f_amplitude, int n_octave_num, float f_persistence)
{
	return perlin_noise.Render_Noise(p_state, p_noise, Vector2f(0, 0),
		f_scale, f_amplitude, n_octave_num, f_persistence);
}

bool Render_Repeatable_Texture(float f_seam_cover_size, float f_crossover_slope)
{
	if(glGetError() != GL_NO_ERROR)
		return false;

	if(!p_fbo->Bind())
		return false;
	// bind FBO

	if(glGetError() != GL_NO_ERROR)
		return false;

	glViewport(0, 0, p_fbo->n_Width(), p_fbo->n_Height());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT);
	// viewport, matrices & clear

	float f_seam_cover_size_h = f_seam_cover_size;
	float f_seam_cover_size_v = f_seam_cover_size;
	if(p_equalized->n_Width() > p_equalized->n_Height())
		f_seam_cover_size_h /= float(p_equalized->n_Width()) / p_equalized->n_Height();
	else if(p_equalized->n_Width() < p_equalized->n_Height())
		f_seam_cover_size_v /= float(p_equalized->n_Height()) / p_equalized->n_Width();
	// seam size for non-square images

	for(int n_pass = 0; n_pass < 2; ++ n_pass) {
		if(!p_fbo->BindColorTexture_2D((n_pass)? *p_temp_tex[1] :
		   *p_temp_tex[0], GL_TEXTURE_2D, 0))
			return false;
		// bind texture

		((n_pass)? p_temp_tex[0] : p_equalized)->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(1);
		p_state->DisableTexture2D();
		p_state->BindProgramObject(0);
		p_state->DisableVertexProgram();
		p_state->DisableFragmentProgram();
		p_state->Color3f(1, 1, 1);

		if(glGetError() != GL_NO_ERROR)
			return false;

		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 1);
		glVertex2f(-1, 1);
		glTexCoord2f(1, 1);
		glVertex2f(1, 1);
		glTexCoord2f(1, 0);
		glVertex2f(1, -1);
		glEnd();
		// draw fullscreen quad (copy source texture)

		p_state->ActiveTextureUnit(1);
		p_noise->Bind_Enable(p_state);
		// set textures

		p_state->EnableBlend();
		p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		p_flaps_shader->t_Find_Parameter("f_crossover_slope").SetValue(f_crossover_slope);

		if(!p_flaps_shader->Bind(p_state))
			return false;

#if 1 //TAKE_FROM_OTHER_SIDE
		if(n_pass) {
			glBegin(GL_QUADS); // horizontal
			glTexCoord3f(1, 0, 0);
			glVertex2f(-1 + 2 * f_seam_cover_size_h, -1);
			glTexCoord3f(1, 1, 0);
			glVertex2f(-1 + 2 * f_seam_cover_size_h, 1);
			glTexCoord3f(1 - f_seam_cover_size_h, 1, 1);
			glVertex2f(-1, 1);
			glTexCoord3f(1 - f_seam_cover_size_h, 0, 1);
			glVertex2f(-1, -1);
			glEnd();
		} else {
			glBegin(GL_QUADS); // vertical
			glTexCoord3f(0, 1 - f_seam_cover_size_v, 1);
			glVertex2f(-1, -1);
			glTexCoord3f(0, 1, 0);
			glVertex2f(-1, -1 + 2 * f_seam_cover_size_v);
			glTexCoord3f(1, 1, 0);
			glVertex2f(1, -1 + 2 * f_seam_cover_size_v);
			glTexCoord3f(1, 1 - f_seam_cover_size_v, 1);
			glVertex2f(1, -1);
			glEnd();
		}
#else // MIRROR (looks bad)
		if(n_pass) {
			glBegin(GL_QUADS);
			glTexCoord3f(1 - f_seam_cover_size, 0, 0);
			glVertex2f(-1 + 2 * f_seam_cover_size, -1);
			glTexCoord3f(1 - f_seam_cover_size, 1, 0);
			glVertex2f(-1 + 2 * f_seam_cover_size, 1);
			glTexCoord3f(1, 1, 1);
			glVertex2f(-1, 1);
			glTexCoord3f(1, 0, 1);
			glVertex2f(-1, -1);
			glEnd();
		} else {
			glBegin(GL_QUADS);
			glTexCoord3f(0, 1, 1);
			glVertex2f(-1, -1);
			glTexCoord3f(0, 1 - f_seam_cover_size, 0);
			glVertex2f(-1, -1 + 2 * f_seam_cover_size);
			glTexCoord3f(1, 1 - f_seam_cover_size, 0);
			glVertex2f(1, -1 + 2 * f_seam_cover_size);
			glTexCoord3f(1, 1, 1);
			glVertex2f(1, -1);
			glEnd();
		}
#endif
		// draw flap to cover seams

		if(n_pass == 1) {
			int n_new_width = p_equalized->n_Width() * (1 - f_seam_cover_size_h);
			int n_new_height = p_equalized->n_Height() * (1 - f_seam_cover_size_v);

			p_state->ActiveTextureUnit(1);
			p_state->DisableTexture2D();
			p_state->ActiveTextureUnit(0);
			delete p_repeatable;
			if(!(p_repeatable = new CGLTexture_2D(p_state,
			   n_new_width, n_new_height, GL_RGBA, true, 0)))
				return false;
			if(!p_repeatable->b_Status())
				return false;
			p_repeatable->Bind_Enable(p_state);
			float f_max_anisotropy;
			glGetFloatv(GL_TEXTURE_MAX_ANISOTROPY_EXT, &f_max_anisotropy);
			p_repeatable->r_Parameters(p_state).Set_Texture_Anisotropy(f_max_anisotropy);
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, n_new_width, n_new_height, 0);
		}

		if(glGetError() != GL_NO_ERROR)
			return false;

		if(!p_fbo->BindColorTexture_2D(0, GL_TEXTURE_2D, 0))
			return false;
		// release texture

		p_state->DisableBlend();
		p_state->ActiveTextureUnit(1);
		p_state->DisableTexture2D();
		p_state->ActiveTextureUnit(0);
	}

	if(!p_fbo->Release())
		return false;
	// release FBO

	return glGetError() == GL_NO_ERROR;
}

CGLTexture_2D *p_CloneInputImageFmt()
{
	CGLTexture_2D *p_texture = new CGLTexture_2D(p_state,
		p_src_image->n_Width(), p_src_image->n_Height(), p_src_image->n_Pixel_Format(), false);
	if(!p_texture)
		return 0;
	if(!p_texture->b_Status()) {
		delete p_texture;
		return 0;
	}
	CGLTextureParams &r_params = p_texture->r_Parameters(p_state);
	r_params.Set_Texture_Min_Filter(GL_NEAREST);
	r_params.Set_Texture_Mag_Filter(GL_NEAREST);
	r_params.Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	r_params.Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	return p_texture;
}

void Cleanup()
{
	for(int i = 0; i < sizeof(p_temp_tex) / sizeof(p_temp_tex[0]); ++ i) {
		if(p_temp_tex[i])
			delete p_temp_tex[i];
	}
	if(p_src_image)
		delete p_src_image;
	if(p_min)
		delete p_min;
	if(p_min_smooth)
		delete p_min_smooth;
	if(p_max)
		delete p_max;
	if(p_max_smooth)
		delete p_max_smooth;
	if(p_equalized)
		delete p_equalized;
	if(p_noise)
		delete p_noise;
	if(p_fbo)
		delete p_fbo;
	if(p_perpix_equalize)
		delete p_perpix_equalize;
	if(p_perpix_grey_equalize)
		delete p_perpix_grey_equalize;
	if(p_flaps_shader)
		delete p_flaps_shader;
	delete p_state;
}

bool CheckOpenGL_Extensions()
{
	if(CGLExtensionHandler::n_GetGL12FuncPointers() ||
	   !CGLExtensionHandler::b_Support_OpenGL(1, 2)) {
		fprintf(stderr, "error: OpenGL 1.2 not supported\n");
		return false; // need 3D texturing
	}
	if(CGLExtensionHandler::n_GetGL13FuncPointers() ||
	   !CGLExtensionHandler::b_Support_OpenGL(1, 3))
		fprintf(stderr, "warning: OpenGL 1.3 not supported\n");
	if(CGLExtensionHandler::n_GetGL14FuncPointers() ||
	   !CGLExtensionHandler::b_Support_OpenGL(1, 4))
		fprintf(stderr, "warning: OpenGL 1.4 not supported\n");
	if(CGLExtensionHandler::n_GetGL15FuncPointers() ||
	   !CGLExtensionHandler::b_Support_OpenGL(1, 5))
		fprintf(stderr, "warning: OpenGL 1.5 not supported\n");
	bool b_have_gl20 = false;
	if(!CGLExtensionHandler::n_GetGL20FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(2, 0))
		b_have_gl20 = CGL_Core_shader::b_Supported();
	else
		fprintf(stderr, "warning: OpenGL 2.0 not supported\n");
	if(CGLExtensionHandler::n_GetGL21FuncPointers() ||
	   !CGLExtensionHandler::b_Support_OpenGL(2, 1)) {
		fprintf(stderr, "warning: OpenGL 2.1 not supported"
			" (have to deal with NVidia render to 3D-texture bug)\n");
	}
	if(!CGL_ARB_shader::b_Supported()) {
		if(!b_have_gl20) {
			fprintf(stderr, "error: GL_ARB_fragment_shader not supported\n");
			return false;
		} else
			fprintf(stderr, "warning: GL_ARB_fragment_shader not supported\n");
	}
	if(!CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_non_power_of_two")) {
		fprintf(stderr, "error: GL_ARB_texture_non_power_of_two not supported\n");
		return false;
	}
	if(!CGLFrameBuffer_FBO::b_Supported()) {
		fprintf(stderr, "error: GL_EXT_framebuffer_object not supported\n");
		return false;
	}
	// check necessary OpenGL extensions

	return true;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("berLame GLUT window");
	// init OpenGL using GLUT

	if(!CheckOpenGL_Extensions()) {
		fprintf(stderr, "error: required OpenGL extensions not found\n");
		return -1;
	}
	// check if all OpenGL extensions are supported

	if(!(p_state = new CGLState(/*wglGetCurrentDC(), wglGetCurrentContext()*/))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create OpenGL state guard

	{
		char p_s_filename[260] = {0};
		OPENFILENAME t_open_file;
		memset(&t_open_file, 0, sizeof(OPENFILENAME));
		t_open_file.lStructSize = sizeof(OPENFILENAME);
		t_open_file.hwndOwner = NULL;
		t_open_file.lpstrFile = p_s_filename;
		t_open_file.nMaxFile = sizeof(p_s_filename);
		t_open_file.lpstrFilter = "Images\0*.jpg;*.jpeg;*.tga\0Jpeg images"
			"\0*.jpg;*.jpeg\0Targa images\0*.tga\0All files\0*.*\0";
		t_open_file.nFilterIndex = 1;
		t_open_file.lpstrFileTitle = NULL;
		t_open_file.nMaxFileTitle = 0;
		t_open_file.lpstrDefExt = ".jpg";
		t_open_file.lpstrInitialDir = NULL;
		t_open_file.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
		t_open_file.lpstrTitle = "Load input image ...";

		if(!GetOpenFileName(&t_open_file)) {
			Cleanup();
			return 0;
		}
		// show open file dialog

		if(!(p_src_image = CTextureLoader::p_LoadTexture(p_state, p_s_filename, false, true))) {
			fprintf(stderr, "error: failed to load \'%s\'\n", p_s_filename);
			Cleanup();
			return -1;
		}
		// load texture from specified file
	}

	CGLTextureParams &r_params = p_src_image->r_Parameters(p_state);
	r_params.Set_Texture_Min_Filter(GL_NEAREST);
	r_params.Set_Texture_Mag_Filter(GL_NEAREST);
	// load input image

	for(int i = 0; i < sizeof(p_temp_tex) / sizeof(p_temp_tex[0]); ++ i) {
		if(!(p_temp_tex[i] = p_CloneInputImageFmt())) {
			fprintf(stderr, "error: not enough memory for textures\n");
			Cleanup();
			return -1;
		}
	}
	if(!(p_min = p_CloneInputImageFmt()) || !(p_min_smooth = p_CloneInputImageFmt()) ||
	   !(p_max = p_CloneInputImageFmt()) || !(p_max_smooth = p_CloneInputImageFmt()) ||
	   !(p_equalized = p_CloneInputImageFmt()) || !(p_repeatable = p_CloneInputImageFmt()) ||
	   !(p_noise = p_CloneInputImageFmt())) {
		fprintf(stderr, "error: not enough memory for textures\n");
		Cleanup();
		return -1;
	}
	// create textures

	if(!(p_fbo = new CGLFrameBuffer_FBO(p_src_image->n_Width(), p_src_image->n_Height(),
	   1, true, 0, false, false, 0, false, false, 0))) {
		fprintf(stderr, "error: error creating FBO\n");
		Cleanup();
		return -1;
	}
	// create framebuffer for rendering to a single texture

	if(!(p_perpix_equalize =
	   CGLImageProc::p_CompileLinkReport(p_state, 0, &t_perpixel_equalize))) {
		fprintf(stderr, "error: error compiling shaders\n");
		Cleanup();
		return -1;
	}
	if(!(p_perpix_grey_equalize =
	   CGLImageProc::p_CompileLinkReport(p_state, 0, &t_perpixel_equalize_mono))) {
		fprintf(stderr, "error: error compiling shaders\n");
		Cleanup();
		return -1;
	}
	if(!(p_flaps_shader = CGLImageProc::p_CompileLinkReport(p_state, 0, &t_flaps_shader))) {
		fprintf(stderr, "error: error compiling shaders\n");
		Cleanup();
		return -1;
	}

	if(!Render_MinMax_Textures(n_min_max_size) ||
	   !Render_SmoothMinMax_Textures(n_gauss_size) ||
	   !Render_Equalized_Texture() ||
	   !Render_Noise_Texture(f_noise_size, f_noise_amp, n_noise_octaves, f_noise_persitance) ||
	   !Render_Repeatable_Texture(f_seam_coverup_size, f_crossover_slope)) {
		fprintf(stderr, "error: Render_*_Textures()\n");
		Cleanup();
		return -1;
	}

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	glutKeyboardFunc(onKeyPressed);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
